'''This module defines valuation model components.'''
from enum import Enum

class VisualizationType(Enum):
    '''A collection of supported visualization types.'''

    LINE = 0
    PIE = 1
    BAR = 2



'''A convenience importing mechanism for all models.'''

from frostaura.models.valuation_result import ValuationResult
from frostaura.models.profit_calculation_result import ProfitCalculationResult
from frostaura.models.visualization_types import VisualizationType
from frostaura.models.symbol_data import SymbolData



'''This module defines calculations engine components.'''
from logging import debug
import pandas as pd
from frostaura.data_access.public_asset_data_access import IPublicAssetDataAccess
from frostaura.engines.asset_calculations_engine import IAssetCalculationsEngine
from frostaura.models import ProfitCalculationResult

class SimpleAssetCalculationsEngine(IAssetCalculationsEngine):
    '''Calculations-related functionality using some maths under-the-hood.'''

    def __init__(self, public_asset_data_access: IPublicAssetDataAccess, config: dict = {}):
        self.public_asset_data_access = public_asset_data_access
        self.config = config

    def __calculate_holdings_ratios__(self, holdings: pd.DataFrame) -> pd.DataFrame:
        '''Determine the ratios that each asset makes up of the overall portfolio adding up to 1.'''

        def determine_usd(row: dict) -> float:
            history: pd.DataFrame = self.public_asset_data_access.get_symbol_history(symbol=row['symbol'])

            return row['shares'] * history.loc[row.name].Close

        holdings_with_usd: pd.DataFrame = holdings.copy()
        holdings_with_usd['usd'] = holdings.apply(determine_usd, axis=1)
        grouped_holdings: pd.DataFrame = holdings_with_usd.groupby(['name', 'symbol'], as_index=False).sum()
        holding_names: list = grouped_holdings['name'].value_counts().index.values
        holding_symbols: list = grouped_holdings['symbol'].value_counts().index.values
        data: dict = {
            'name': holding_names,
            'symbol': holding_symbols,
            'ratio': list()
        }

        for row_index, row in grouped_holdings.iterrows():
            data['ratio'].append(row['usd'] / grouped_holdings['usd'].sum())

        return pd.DataFrame(data)

    def calculate_holdings_profits(self, holdings: pd.DataFrame) -> pd.DataFrame:
        '''Determine individual asset profit ratio & profit USD and interpolate them into a copy of the given holdings.'''

        holding_names: list = holdings['name'].value_counts().index.values
        holding_symbols: list = holdings['symbol'].value_counts().index.values
        data: dict = {
            'name': holding_names,
            'symbol': holding_symbols,
            'total_purchased_usd': list(),
            'total_purchased_shares': list(),
            'total_current_usd': list(),
            'total_profit_ratio': list(),
            'total_profit_usd': list()
        }

        for holding_symbol in holding_symbols:
            debug(f'Calculating profit for asset "{holding_symbol}".')

            transactions_by_date_asc: pd.DataFrame = holdings \
                                                        .loc[holdings['symbol'] == holding_symbol] \
                                                        .sort_index(ascending=True)
            history: pd.DataFrame = self.public_asset_data_access.get_symbol_history(symbol=holding_symbol)
            total_purchased_usd: float = 0
            total_purchased_shares: float = 0

            for transaction_date, row in transactions_by_date_asc.iterrows():
                transaction_value: float = row['shares']
                transaction_close: float = history.loc[transaction_date]['Close']

                debug(f'[{holding_symbol}] Processing transaction value ${transaction_value} on {transaction_date}.')

                total_purchased_usd += transaction_close * transaction_value
                total_purchased_shares += transaction_value

            total_current_usd: float = history.iloc[-1]['Close'] * total_purchased_shares

            if total_current_usd > 1:
                data['total_purchased_usd'].append(total_purchased_usd)
                data['total_purchased_shares'].append(total_purchased_shares)
                data['total_current_usd'].append(total_current_usd)
                data['total_profit_ratio'].append((1 - min(total_purchased_usd, total_current_usd) / max(total_purchased_usd, total_current_usd)) * 100)
                data['total_profit_usd'].append(total_current_usd - total_purchased_usd)
            else:
                data['total_purchased_usd'].append(0)
                data['total_purchased_shares'].append(0)
                data['total_current_usd'].append(0)
                data['total_profit_ratio'].append(0)
                data['total_profit_usd'].append(0)

        ratios: pd.DataFrame = self.__calculate_holdings_ratios__(holdings=holdings)

        return pd.DataFrame(data).set_index('symbol').join(other=ratios.drop(columns=['name']).set_index('symbol')).reset_index()

    def calculate_holdings_profit(self, holdings: pd.DataFrame) -> ProfitCalculationResult:
        '''Determine the holdings profit percentage & profit USD, given the holdings.'''

        debug('Calculating overall holdings profits.')

        holdings_profits: pd.DataFrame = self.calculate_holdings_profits(holdings=holdings)
        total_purchased_usd: float = holdings_profits['total_purchased_usd'].sum()
        total_current_usd: float = holdings_profits['total_current_usd'].sum()
        total_profit_ratio: float = (1 - min(total_purchased_usd, total_current_usd) / max(total_purchased_usd, total_current_usd)) * 100
        total_profit_usd: float = total_current_usd - total_purchased_usd

        return ProfitCalculationResult(percentage=total_profit_ratio, value=total_profit_usd)



'''This module defines the interface for the notifications data access components.'''
from telegram import ParseMode, Update
from telegram.ext import Updater, MessageHandler, Filters
import io
import matplotlib.pyplot as plt
import prettytable as pt
import pandas as pd
from frostaura.data_access.notifications_data_access import INotificationsDataAccess

class TelegramNotificationsDataAccess(INotificationsDataAccess):
    '''Component to perform resource related actions.'''

    def __init__(self, bot_token: str, chat_id: str, message_handler = None):
        assert bot_token is not None

        self.bot_token = bot_token

        self.updater = Updater(token=self.bot_token)
        self.dispatcher = self.updater.dispatcher
        self.bot = self.updater.bot
        self.chat_id = chat_id
        self.message_handler = message_handler

        self.dispatcher.add_handler(MessageHandler(Filters.all, self.__on_message__))
        self.updater.start_polling()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, t_b):
        self.updater.stop()
        self.updater = None
        self.dispatcher = None
        self.bot = None

    def __get_figure_buffer__(self, figure: plt.Figure) -> io.BytesIO:
        buffer = io.BytesIO()
        figure.savefig(buffer, format='png')
        buffer.seek(0)

        return buffer

    def __escape_for_markdown__(self, text: str) -> str:
        return text \
            .replace("_", "\_") \
            .replace("[", "\[") \
            .replace("]", "\]") \
            .replace("(", "\(") \
            .replace(")", "\)") \
            .replace("~", "\~") \
            .replace(">", "\>") \
            .replace("#", "\#") \
            .replace("+", "\+") \
            .replace("-", "\-") \
            .replace("=", "\=") \
            .replace("|", "\|") \
            .replace("{", "\{") \
            .replace("}", "\}") \
            .replace(".", "\.") \
            .replace("!", "\!")

    def __on_message__(self, update: Update, context) -> None:
        if self.message_handler is not None:
            self.message_handler(update, context)

    def send_text(self, text: str):
        '''Send a text notification as markdown.'''

        self.bot.send_message(chat_id=self.chat_id, text=self.__escape_for_markdown__(text), parse_mode=ParseMode.MARKDOWN_V2)

    def send_figure(self, figure: plt.Figure):
        self.bot.send_photo(chat_id=self.chat_id, photo=self.__get_figure_buffer__(figure))

    def send_dataframe(self, dataframe: pd.DataFrame):
        table = pt.PrettyTable(tuple(dataframe.columns))
        
        for index, row in dataframe.iterrows():
            row_data = list()

            for column_name in dataframe.columns:
                row_data.append(row[column_name])
            
            table.add_row(row_data)

        self.bot.send_message(chat_id=self.chat_id, text=f'<pre>{table}</pre>', parse_mode=ParseMode.HTML)



'''This module defines visualization engine components.'''
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import seaborn as sns
from frostaura.engines.visualization_engine import IVisualizationEngine
from frostaura.models.visualization_types import VisualizationType

class FrostAuraVisualizationEngine(IVisualizationEngine):
    '''Component to perform functions related to visualization that is themed for FrostAura.'''

    def get_figure(self,
                   data: pd.DataFrame,
                   x: str,
                   y: str,
                   graph_type: VisualizationType,
                   title: str=None,
                   subtitle: str=None,
                   legend: bool=False,
                   line_label: str=None,
                   x_tick_format_str: str=None,
                   y_tick_format_str: str=None) -> list:
        '''Generate a visualization for the given parameters and return a matplotlib figure.'''

        fa_light: str = '#007bff'
        fa_dark: str = '#00379b'
        fa_gray: str = '#707070'
        fa_white: str = 'white'

        line_width: float = 1
        grid_line_width: float = 0.5
        outer_background_color: str = fa_white
        inner_background_color: str = fa_white
        labels_color: str = fa_gray
        grid_frame_color: str = fa_white
        should_show_grid: bool = True
        grid_color: str = fa_light
        title_legend_color: str = fa_dark
        x_y_colors: str = fa_light

        sns.set(font_scale=0.75, rc={
            'axes.facecolor': inner_background_color,
            'figure.facecolor': outer_background_color,
            'axes.labelcolor': labels_color,
            'axes.edgecolor': grid_frame_color,
            'axes.grid': should_show_grid,
            'grid.color': grid_color,
            'grid.linestyle': '--',
            'text.color': title_legend_color,
            'xtick.color': x_y_colors,
            'ytick.color': x_y_colors,
            'lines.linewidth': line_width,
            'grid.linewidth': grid_line_width
        })
        figsize = (4,4) if graph_type == VisualizationType.PIE else (8, 4)
        fig, axs = plt.subplots(1, dpi=600, figsize=figsize)

        if line_label is not None and y_tick_format_str is not None:
            if not type(line_label) == str:
                line_label = y_tick_format_str.format(x=line_label)

        if graph_type == VisualizationType.LINE:
            sns.lineplot(x=x, y=y, data=data, ci=False, color=fa_dark, ax=axs, label=line_label)
        elif graph_type == VisualizationType.BAR:
            raise NotImplementedError('Bar plots have outstanding styling issues and is not yet supported.')
            #sns.barplot(x=x, y=y, data=data, ci=False, color=fa_dark, ax=axs, label=line_label)
        elif graph_type == VisualizationType.PIE:
            labels: list = data[x].values
            values: list = data[y].values

            plt.pie(x=values, labels=labels)
        else:
            raise NotImplementedError(f'The figure type "{graph_type}" is not supported.')

        if title is not None:
            mid = (fig.subplotpars.right + fig.subplotpars.left)/2
            fig.suptitle(title, x=mid)

        if subtitle is not None:
            axs.set_title(subtitle)

        if y_tick_format_str is not None:
            tick = mtick.StrMethodFormatter(y_tick_format_str)
            axs.yaxis.set_major_formatter(tick)

        if x_tick_format_str is not None:
            tick = mtick.StrMethodFormatter(x_tick_format_str)
            axs.xaxis.set_major_formatter(tick)

        current_legend = axs.get_legend()

        if current_legend is not None:
            current_legend.remove()

        if legend:
            fig.legend(numpoints=1)

        return fig, axs



'''This module defines public asset data access components.'''
import pandas as pd
from frostaura.models.symbol_data import SymbolData

class IPublicAssetDataAccess:
    '''Component to perform functions related to public assets.'''

    def get_symbol_history(self, symbol: str) -> pd.DataFrame:
        '''Get historical price movements for a given symbol.'''

        raise NotImplementedError()

    def get_symbol_info(self, symbol: str) -> SymbolData:
        '''Get the data for a specific symbol.'''

        raise NotImplementedError()

    def augment_symbols_info(self, symbols: pd.DataFrame) -> pd.DataFrame:
        '''Add info to a daraframe containing symbols.'''

        raise NotImplementedError()



'''This module defines visualization engine components.'''
import pandas as pd
import matplotlib.pyplot as plt
from frostaura.models.visualization_types import VisualizationType

class IVisualizationEngine:
    '''Component to perform functions related to visualization.'''

    def get_figure(self,
                   data: pd.DataFrame,
                   x: str,
                   y: str,
                   graph_type: VisualizationType,
                   title: str,
                   subtitle: str,
                   legend: bool,
                   line_label: str,
                   x_tick_format_str: str,
                   y_tick_format_str: str) -> list:
        '''Generate a visualization for the given parameters and return a matplotlib figure.'''

        raise NotImplementedError()



'''This module defines private manager components.'''

class IAssetReportingManager:
    '''Component to perform functions related to asset reporting.'''

    def send_reports(self):
        '''Generate and send asset reports.'''

        raise NotImplementedError()



from setuptools import setup, find_packages
from pathlib import Path

ROOT_DIR = Path(__file__).parent
VERSION = '1.0.35'
DESCRIPTION = 'FrostAura provides a range of open-source components provide a variety of problem domains.'
LONG_DESCRIPTION = DESCRIPTION #(ROOT_DIR / '../README.md').read_text()

setup(
    name='frostaura',
    version=VERSION,
    author='Dean Martin',
    author_email='dean.martin@frostaura.net',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/faGH/fa.intelligence.notebooks',
    packages=find_packages(),
    install_requires=[],
    keywords=[
        'frostaura',
        'deep learning',
        'machine learning',
        'market analysis',
        'bots',
        'visualization',
        'data exploration',
        'data analysis',
        'statistics',
        'neural networks',
        'securities',
        'stocks',
        'assets'
    ],
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.7'
)



'''This module defines embedded resources data access component.'''
from typing import BinaryIO
from logging import info
import inspect
from frostaura.data_access.resources_data_access import IResourcesDataAccess

class EmbeddedResourcesDataAccess(IResourcesDataAccess):
    '''Component to perform embedded resource related actions.'''

    def __init__(self, config: dict = {}):
        self.config = config

    def get_resource(self, path: str) -> BinaryIO:
        '''Get a resource as a byte stream that was embedded in a given package.'''

        directories_to_move_up: int = 2
        current_executing_path: str = inspect.getfile(self.__class__)
        current_executing_path = current_executing_path.replace('\\', '/')
        root_path: str = '/'.join(current_executing_path.split('/')[:-directories_to_move_up])
        resource_path: str = f'{root_path}/resources/{path}'

        info(f'Fetching embedded resource "{resource_path}".')

        return open(resource_path)



'''This module defines valuation model components.'''

class ValuationResult:
    '''A model with all the post-valuation data for an asset.'''

    def __init__(self,
                 symbol: str,
                 company_name: str,
                 current_price: float,
                 valuation_method: str,
                 valuation_price: float,
                 margin_of_safety: float):
        assert margin_of_safety > 0 and margin_of_safety < 1

        self.symbol = symbol
        self.company_name = company_name
        self.current_price = current_price
        self.valuation_method = valuation_method
        self.valuation_price = valuation_price
        self.fair_price = valuation_price * (1 - margin_of_safety)
        self.absolute_current_v_valuation_delta = 1 - (min(self.fair_price, current_price) / max(self.fair_price, current_price))
        self.is_overvalued = self.fair_price < current_price
        self.margin_of_safety = margin_of_safety
        self.divident_payout_frequency_in_months = 0 # TODO: Fetch this data.
        self.annual_dividend_percentage = 0 # TODO: Fetch this data.



'''This module definessymbol data model components.'''

class SymbolData:
    '''A model with all the detailed information for a symbol.'''

    def __init__(self,
                 symbol: str,
                 company_name: str,
                 current_price: float,
                 eps: float,
                 annual_growth_projected: float,
                 current_yield_of_aaa_corporate_bonds: float = 4.27, # https://fred.stlouisfed.org/series/AAA
                 pe_base_non_growth_company: float = 8.5,
                 average_yield_of_aaa_corporate_bonds: float = 4.4):
        self.symbol = symbol
        self.company_name = company_name
        self.current_price = current_price
        self.eps = eps
        self.annual_growth_projected = annual_growth_projected
        self.current_yield_of_aaa_corporate_bonds = current_yield_of_aaa_corporate_bonds
        self.pe_base_non_growth_company = pe_base_non_growth_company
        self.average_yield_of_aaa_corporate_bonds = average_yield_of_aaa_corporate_bonds
